// -*- Mode: C++; indent-tabs-mode: nil; tab-width: 2 -*-
/*
 * Copyright (C) 2011 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Jason Smith <jason.smith@canonical.com>
 */

#ifndef SWITCHERVIEW_H
#define SWITCHERVIEW_H

#include <tuple>

#include "SwitcherModel.h"
#include "unity-shared/AbstractIconRenderer.h"
#include "unity-shared/StaticCairoText.h"
#include "unity-shared/LayoutSystem.h"
#include "unity-shared/BackgroundEffectHelper.h"
#include "unity-shared/UnityWindowView.h"

#include <Nux/View.h>
#include <NuxCore/Property.h>

#include <UnityCore/GLibSource.h>


namespace unity
{
namespace launcher
{
class AbstractLauncherIcon;
}
namespace switcher
{
class View
{
  public:

    typedef sigc::slot<void, int, int, int, int> MouseButtonCallback;
    typedef sigc::slot <void, int, int, int, int, unsigned int, unsigned int> MouseDragCallback;

    typedef std::tuple <sigc::connection,
                        sigc::connection,
                        sigc::connection> MouseEventConnections;

    class Impl;
    typedef std::unique_ptr<Impl> ImplPtr;
    typedef std::shared_ptr<View> Ptr;
    typedef std::function<ImplPtr()> ViewFactoryFunc;

    View(ViewFactoryFunc const&);

    ui::LayoutWindow::Vector ExternalTargets();

    void SetModel(SwitcherModel::Ptr model);
    SwitcherModel::Ptr GetModel();

    MouseEventConnections ConnectToMouseEvents(const MouseButtonCallback &mouse_down_callback,
                                               const MouseButtonCallback &mouse_up_callback,
                                               const MouseDragCallback   &mouse_drag_callback);
    void QueueDraw();
    nux::Geometry GetAbsoluteGeometry() const;

    // Returns the index of the icon at the given position, in window coordinates.
    // If there's no icon there, -1 is returned.
    int IconIndexAt(int x, int y);

    // Property setters
    void SetBackgroundColor(nux::Color const& color);
    void SetMonitor(unsigned int monitor);
    void SetBorderSize(unsigned int);
    void SetFlatSpacing(int);
    void SetTextSize(unsigned int);
    void SetIconSize(unsigned int);
    void SetTileSize(unsigned int);
    void SetAnimationLength(unsigned int);
    void SetRenderBoxes(bool);

    // Property getters
    unsigned int GetBorderSize();
    unsigned int GetFlatSpacing();
    unsigned int GetTextSize();
    unsigned int GetIconSize();
    unsigned int GetTileSize();
    unsigned int GetAnimationLength();

    void SetupBackground(bool enabled = true);

  private:
    std::unique_ptr<Impl> impl_;
};

class View::Impl
{
public:

  typedef switcher::View::MouseButtonCallback MouseButtonCallback;
  typedef switcher::View::MouseDragCallback MouseDragCallback;
  typedef switcher::View::MouseEventConnections MouseEventConnections;

  Impl();

  virtual ~Impl() {};

  virtual ui::LayoutWindow::Vector ExternalTargets() = 0;

  virtual void SetModel(SwitcherModel::Ptr model) = 0;
  virtual SwitcherModel::Ptr GetModel() = 0;

  virtual MouseEventConnections ConnectToMouseEvents(const MouseButtonCallback &mouse_down_callback,
                                                     const MouseButtonCallback &mouse_up_callback,
                                                     const MouseDragCallback   &mouse_drag_callback) = 0;
  virtual void QueueDraw() = 0;
  virtual nux::Geometry GetAbsoluteGeometry() const = 0;

  // Returns the index of the icon at the given position, in window coordinates.
  // If there's no icon there, -1 is returned.
  virtual int IconIndexAt(int x, int y) = 0;

  // Property setters
  virtual void SetBackgroundColor(nux::Color const& color) = 0;
  virtual void SetMonitor(unsigned int monitor) = 0;
  virtual void SetBorderSize(unsigned int) = 0;
  virtual void SetFlatSpacing(int) = 0;
  virtual void SetTextSize(unsigned int) = 0;
  virtual void SetIconSize(unsigned int) = 0;
  virtual void SetTileSize(unsigned int) = 0;
  virtual void SetAnimationLength(unsigned int) = 0;
  virtual void SetRenderBoxes(bool) = 0;

  // Property getters
  virtual unsigned int GetBorderSize() = 0;
  virtual unsigned int GetFlatSpacing() = 0;
  virtual unsigned int GetTextSize() = 0;
  virtual unsigned int GetIconSize() = 0;
  virtual unsigned int GetTileSize() = 0;
  virtual unsigned int GetAnimationLength() = 0;
  virtual void SetupBackground(bool enabled) = 0;
};

class NuxView : public ui::UnityWindowView,
                public unity::switcher::View::Impl
{
  NUX_DECLARE_OBJECT_TYPE(NuxView, ui::UnityWindowView);
public:

  NuxView();

  ui::LayoutWindow::Vector ExternalTargets();

  void SetModel(SwitcherModel::Ptr model);
  SwitcherModel::Ptr GetModel();

  // Returns the index of the icon at the given position, in window coordinates.
  // If there's no icon there, -1 is returned.
  int IconIndexAt(int x, int y);

  MouseEventConnections ConnectToMouseEvents(MouseButtonCallback const& mouse_down_callback,
                                             MouseButtonCallback const& mouse_up_callback,
                                             MouseDragCallback const&   mouse_drag_callback);

protected:
  // Introspectable methods
  void SetupBackground(bool enabled);
  std::string GetName() const;
  void AddProperties(GVariantBuilder* builder);

  void PreDraw(nux::GraphicsEngine& GfxContext, bool force_draw);
  void DrawOverlay(nux::GraphicsEngine& GfxContext, bool force_draw, nux::Geometry clip);
  void QueueDraw();
  nux::Geometry GetAbsoluteGeometry() const;
  nux::Geometry GetBackgroundGeometry();

  ui::RenderArg InterpolateRenderArgs(ui::RenderArg const& start, ui::RenderArg const& end, float progress);
  nux::Geometry InterpolateBackground(nux::Geometry const& start, nux::Geometry const& end, float progress);

  std::list<ui::RenderArg> RenderArgsFlat(nux::Geometry& background_geo, int selection, timespec const& current);

  ui::RenderArg CreateBaseArgForIcon(launcher::AbstractLauncherIcon::Ptr const& icon);
private:
  void OnSelectionChanged(launcher::AbstractLauncherIcon::Ptr const& selection);
  void OnDetailSelectionChanged(bool detail);
  void OnDetailSelectionIndexChanged(unsigned int index);

  void OnIconSizeChanged(int size);
  void OnTileSizeChanged(int size);

  nux::Geometry UpdateRenderTargets(nux::Point const& center, timespec const& current);
  void OffsetRenderTargets(int x, int y);

  nux::Size SpreadSize();

  void GetFlatIconPositions(int n_flat_icons,
                            int size,
                            int selection,
                            int &first_flat,
                            int &last_flat,
                            int &half_fold_left,
                            int &half_fold_right);

  void SaveLast();

  // Property setters
  void SetBackgroundColor(nux::Color const& color);
  void SetMonitor(unsigned int monitor);
  void SetBorderSize(unsigned int);
  void SetFlatSpacing(int);
  void SetTextSize(unsigned int);
  void SetIconSize(unsigned int);
  void SetTileSize(unsigned int);
  void SetAnimationLength(unsigned int);
  void SetRenderBoxes(bool);

  // Property getters
  unsigned int GetBorderSize();
  unsigned int GetFlatSpacing();
  unsigned int GetTextSize();
  unsigned int GetIconSize();
  unsigned int GetTileSize();
  unsigned int GetAnimationLength();

  nux::Property<bool> render_boxes;
  nux::Property<int> border_size;
  nux::Property<int> flat_spacing;
  nux::Property<int> icon_size;
  nux::Property<int> minimum_spacing;
  nux::Property<int> tile_size;
  nux::Property<int> vertical_size;
  nux::Property<int> text_size;
  nux::Property<int> animation_length;
  nux::Property<int> monitor;
  nux::Property<double> spread_size;

  SwitcherModel::Ptr model_;
  ui::LayoutSystem layout_system_;
  ui::AbstractIconRenderer::Ptr icon_renderer_;
  nux::ObjectPtr<StaticCairoText> text_view_;

  bool animation_draw_;
  bool target_sizes_set_;


  std::list<ui::RenderArg> last_args_;
  std::list<ui::RenderArg> saved_args_;

  nux::Geometry last_background_;
  nux::Geometry saved_background_;

  ui::LayoutWindow::Vector render_targets_;

  timespec current_;
  timespec save_time_;

  glib::Source::UniquePtr redraw_idle_;
};

class NuxViewWrapper : public unity::switcher::View::Impl
{
  public:

    NuxViewWrapper(nux::ObjectPtr<NuxView> const& nuxView);

    ui::LayoutWindow::Vector ExternalTargets();

    void SetModel(SwitcherModel::Ptr model);
    SwitcherModel::Ptr GetModel();

    MouseEventConnections ConnectToMouseEvents(const MouseButtonCallback &mouse_down_callback,
                                               const MouseButtonCallback &mouse_up_callback,
                                               const MouseDragCallback   &mouse_drag_callback);
    void QueueDraw();
    nux::Geometry GetAbsoluteGeometry() const;

    // Returns the index of the icon at the given position, in window coordinates.
    // If there's no icon there, -1 is returned.
    int IconIndexAt(int x, int y);

    // Property setters
    void SetBackgroundColor(nux::Color const& color);
    void SetMonitor(unsigned int monitor);
    void SetBorderSize(unsigned int);
    void SetFlatSpacing(int);
    void SetTextSize(unsigned int);
    void SetIconSize(unsigned int);
    void SetTileSize(unsigned int);
    void SetAnimationLength(unsigned int);
    void SetRenderBoxes(bool);

    // Property getters
    unsigned int GetBorderSize();
    unsigned int GetFlatSpacing();
    unsigned int GetTextSize();
    unsigned int GetIconSize();
    unsigned int GetTileSize();
    unsigned int GetAnimationLength();

    void SetupBackground(bool enabled);

    nux::Area * GetTopLevelViewWindow();
  private:
    nux::ObjectPtr<nux::Object> reference_;
    unity::switcher::View::Impl *impl_;
};
}
}

#endif // SWITCHERVIEW_H

